@@ -1,5 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 1999-2019 Free Software Foundation, Inc.
+   Copyright (C) 1999-2020 Free Software Foundation, Inc.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -55,9 +55,7 @@ struct _amd_message
   mu_off_t body_start;      /* Offset of body start in the message file */
   mu_off_t body_end;        /* Offset of body end (size of file, effectively)*/
 
-  int orig_flags;           /* Original attribute flags */
   int attr_flags;           /* Current attribute flags */
-	
 
   time_t mtime;             /* Time of last modification */
   size_t header_lines;      /* Number of lines in the header part */
@@ -69,7 +67,8 @@ struct _amd_message
 
 /* AMD capabilities */
 #define MU_AMD_STATUS    0x01  /* format keeps status flags */
-#define MU_AMD_IMAPBASE  0x02  /* format keeps IMAP base */
+
+#define MU_AMD_PROP      0x04  /* prop file existed */
 
 struct _amd_data
 {
@@ -79,7 +78,7 @@ struct _amd_data
   int (*msg_finish_delivery) (struct _amd_data *, struct _amd_message *,
 			      const mu_message_t);
   void (*msg_free) (struct _amd_message *);
-  int (*cur_msg_file_name) (struct _amd_message *, char **);	
+  int (*cur_msg_file_name) (struct _amd_message *, int, char **);	
   int (*new_msg_file_name) (struct _amd_message *, int, int, char **);
   int (*scan0)     (mu_mailbox_t mailbox, size_t msgno, size_t *pcount,
 		    int do_notify);
@@ -129,5 +128,6 @@ struct _amd_message *_amd_get_message (struct _amd_data *amd, size_t msgno);
 int amd_msg_lookup (struct _amd_data *amd, struct _amd_message *msg,
 		    size_t *pret);
 int amd_remove_dir (const char *name);
+void amd_reset_uidvalidity (struct _amd_data *amd);
 
 #endif		    
