@@ -29,6 +29,7 @@
 #include <mailutils/filter.h>
 #include <mailutils/cctype.h>
 #include <mailutils/datetime.h>
+#include <mailutils/sys/message.h>
 
 #define ATTRIBUTE_IS_DELETED(flag)        (flag & MU_ATTRIBUTE_DELETED)
 #define ATTRIBUTE_IS_EQUAL(flag1, flag2)  (flag1 == flag2)
@@ -82,11 +83,11 @@ mbox_open (mu_mailbox_t mailbox, int flags)
   /* Get a stream.  */
   if (mailbox->stream == NULL)
     {
-      /* We do not try to mmap for CREAT or APPEND, it is not supported.  */
-      status = (flags & MU_STREAM_CREAT)
-	          || (mailbox->flags & MU_STREAM_APPEND);
-
-      /* Try to mmap () the file first.  */
+      /* Prefer mmap, unless the mailbox file does not exist and CREAT is
+	 requested or the mailbox is being opened for appending */
+      status = (((flags & MU_STREAM_CREAT) && access (mud->name, F_OK)) ||
+		(mailbox->flags & MU_STREAM_APPEND));
+      
       if (status == 0)
 	{
 	  status = mu_mapfile_stream_create (&mailbox->stream, mud->name,
@@ -118,7 +119,8 @@ mbox_close (mu_mailbox_t mailbox)
 {
   mbox_data_t mud = mailbox->data;
   size_t i; 
-
+  int rc;
+  
   if (mud == NULL)
     return EINVAL;
 
@@ -159,7 +161,10 @@ mbox_close (mu_mailbox_t mailbox)
   mud->uidnext = 0;
   mu_monitor_unlock (mailbox->monitor);
 
-  return mu_stream_close (mailbox->stream);
+  rc = mu_stream_close (mailbox->stream);
+  if (rc == 0)
+    mu_stream_destroy (&mailbox->stream);
+  return rc;
 }
 
 static int
@@ -519,6 +524,13 @@ _msg_attribute_setup (mu_message_t msg, mbox_message_t mum)
     }
   return status;
 }
+
+static void
+mbox_message_detach (mu_message_t msg)
+{
+  mbox_message_t mum = mu_message_get_owner (msg);
+  mum->message = NULL;
+}
 
 /* Initialize a message */
 static int
@@ -558,6 +570,8 @@ new_message (mu_mailbox_t mailbox, mbox_message_t mum, mu_message_t *pmsg)
   if (status != 0)
     return status;
 
+  msg->_detach = mbox_message_detach;
+    
   status = _msg_setup (msg, mum);
   if (status)
     {
@@ -769,7 +783,7 @@ mbox_message_unseen (mu_mailbox_t mailbox, size_t *pmsgno)
 }
 
 static int
-mbox_uidvalidity (mu_mailbox_t mailbox, unsigned long *puidvalidity)
+mbox_get_uidvalidity (mu_mailbox_t mailbox, unsigned long *puidvalidity)
 {
   mbox_data_t mud = mailbox->data;
   int status = mbox_messages_count (mailbox, NULL);
@@ -787,6 +801,24 @@ mbox_uidvalidity (mu_mailbox_t mailbox, unsigned long *puidvalidity)
   return 0;
 }
 
+static int
+mbox_set_uidvalidity (mu_mailbox_t mailbox, unsigned long uidvalidity)
+{
+  mbox_data_t mud = mailbox->data;
+  int status = mbox_messages_count (mailbox, NULL);
+  if (status != 0)
+    return status;
+  /* If we did not start a scanning yet do it now.  */
+  if (mud->messages_count == 0)
+    {
+      status = mbox_scan0 (mailbox, 1, NULL, 0);
+      if (status != 0)
+	return status;
+    }
+  mud->uidvalidity = uidvalidity;
+  return 0;
+}
+
 static int
 mbox_uidnext (mu_mailbox_t mailbox, size_t *puidnext)
 {
@@ -1058,14 +1090,12 @@ append_message_to_stream (mu_stream_t ostr, mu_message_t msg,
 
       if (flags & MBOX_FIRSTMSG)
 	{
-	  /* FIXME: Perhaps printf should return error code,
-	     like the rest of stream functions. */
-	  mu_stream_printf (ostr, "%s: %lu %u\n",
-			    MU_HEADER_X_IMAPBASE,
-			    (unsigned long) mud->uidvalidity,
-			    (unsigned) mud->uidnext);
-	  if (mu_stream_err (ostr))
-	    return mu_stream_last_error (ostr);
+	  status = mu_stream_printf (ostr, "%s: %lu %u\n",
+				     MU_HEADER_X_IMAPBASE,
+				     (unsigned long) mud->uidvalidity,
+				     (unsigned) mud->uidnext);
+	  if (status)
+	    return status;
 	}
 
       status = msg_attr_to_stream (ostr, msg);
@@ -1099,10 +1129,10 @@ append_message_to_stream (mu_stream_t ostr, mu_message_t msg,
   mu_stream_unref (istr);
   if (status == 0)
     {
-      status = mu_stream_copy (ostr, flt, 0, NULL);
+      status = mu_stream_copy_nl (ostr, flt, 0, NULL);
       mu_stream_destroy (&flt);
       if (status == 0)
-      	status = mu_stream_write (ostr, "\n", 1, NULL);
+	status = mu_stream_write (ostr, "\n", 1, NULL);
     }
   
   return status;
@@ -1130,9 +1160,8 @@ mbox_append_message (mu_mailbox_t mailbox, mu_message_t msg)
     }
 
   status = mu_stream_seek (mailbox->stream, 0, MU_SEEK_END, &size);
-  if (status)
-    return status;
-  status = append_message_to_stream (mailbox->stream, msg, mud, 0);
+  if (status == 0)
+    status = append_message_to_stream (mailbox->stream, msg, mud, 0);
   if (mailbox->locker)
     mu_locker_unlock (mailbox->locker);
 
@@ -1270,7 +1299,7 @@ mbox_expunge_unlocked (mu_mailbox_t mailbox, size_t dirty, int remove_deleted,
 	    }
 	  status = mu_stream_copy (tempstr, mailbox->stream,
 				   mum->body_end - mum->envel_from, NULL);
-	  /* Terminate messate with an empty line */
+	  /* Terminate message with an empty line */
 	  if (status == 0)
 	    status = mu_stream_write (tempstr, "\n", 1, NULL);      
 	  if (status)
@@ -1426,7 +1455,7 @@ mbox_expunge0 (mu_mailbox_t mailbox, int remove_deleted)
       (status = mu_locker_lock (mailbox->locker)) != 0)
     return status;
 
-  status = mu_temp_file_stream_create (&tempstr, NULL, 0);
+  status = mu_temp_stream_create (&tempstr, 0);
   if (status == 0)
     {
       sigset_t signalset;
@@ -1527,7 +1556,8 @@ _mailbox_mbox_init (mu_mailbox_t mailbox)
   mailbox->_message_unseen = mbox_message_unseen;
   mailbox->_expunge = mbox_expunge;
   mailbox->_sync = mbox_sync;
-  mailbox->_uidvalidity = mbox_uidvalidity;
+  mailbox->_get_uidvalidity = mbox_get_uidvalidity;
+  mailbox->_set_uidvalidity = mbox_set_uidvalidity;
   mailbox->_uidnext = mbox_uidnext;
   mailbox->_quick_get_message = mbox_quick_get_message;
 
